'''
options: #选项
  repo_root: repo         # submodule保存根目录
  push: true              # 同步后是否push
  pr: true                # 是否创建pr，需要目标仓库配置token和type
  proxy_fix: true         # 是否将https代理改成http://开头，解决开着代理情况下，无法发起https请求的问题
  use_system_proxy: false  # 是否使用系统代理
'''
from lib.util import merge_from_dict


class Options:
    repo_root: str = 'repo'
    push: bool = True
    pull_request: bool = True
    proxy_fix: bool = True
    use_system_proxy: bool = True

    def __init__(self, conf_dict=None):
        if conf_dict is None:
            conf_dict = {}
        merge_from_dict(self, conf_dict)



import datetime
import json
import os
import shutil
import time

import git

from lib.api.index import api_clients
from lib.handler.helper import save_work_repo
from lib.http import Http
from lib.logger import logger
from lib.model.config import Config
from lib.model.sync import SyncTask
from lib.util import shell, get_dict_value, check_need_push, set_dict_value, is_blank_dir
from lib.util_git import force_checkout_main_branch, checkout_branch, collection_commit_message, \
    get_git_modify_file_count


def read_status(root):
    file_path = f'{root}/status.json'
    if not os.path.exists(file_path):
        return {}
    fo = open(file_path, "r")
    config_str = fo.read()
    fo.close()
    if config_str is None:
        return {}
    try:
        return json.loads(config_str)
    except Exception as e:
        print(e)
        return {}


def save_status(root, status):
    # 创建配置文件
    file_path = f'{root}/status.json'
    # 写入配置文件
    config_str = json.dumps(status)
    fo = open(file_path, "w")
    fo.write(config_str)
    fo.close()
    return status


def text_center(text: str, length=40):
    return text.center(length, '-')


def sync_func(task: SyncTask, src_dir, target_dir):
    if task.copy_script is None or task.copy_script.strip() == '':
        shutil.copytree(src_dir, target_dir)
    else:
        if len(task.copy_script.splitlines()) > 1:
            # 多行 则表示是脚本，直接执行
            exec(task.copy_script)
        else:
            # 单行，表示是文件，加载文件模块，并执行copy方法
            filepath = os.path.abspath(task.copy_script)
            do_copy_from_file_module(filepath, task, src_dir, target_dir)


def do_copy_from_file_module(filepath, task: SyncTask, src_dir, target_dir):
    # 加载文件模块，并执行copy方法
    import importlib.util
    import sys
    # For illustrative purposes.
    import tokenize
    file_path = tokenize.__file__
    module_name = tokenize.__name__

    spec = importlib.util.spec_from_file_location('copy', filepath)
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    module.copy(task, src_dir, target_dir)


class SyncHandler:

    def __init__(self, work_root, config):
        self.work_root = work_root
        self.config: Config = config
        self.status = read_status(work_root)
        self.conf_repo = config.repo
        self.conf_options = config.options
        self.conf_repo_root = self.conf_options.repo_root

        proxy_fix = self.conf_options.proxy_fix
        use_system_proxy = self.conf_options.use_system_proxy
        self.http = Http(use_system_proxy=use_system_proxy, proxy_fix=proxy_fix)

        self.repo: git.Repo = git.Repo.init(path=work_root)

    def handle(self):
        """
        处理 sync 命令
        """
        logger.info(text_center("sync start"))
        config = self.config
        os.chdir(self.work_root)
        # save_work_repo(self.repo, "save work repo before sync", self.config.options.push)
        is_init = False
        ref_count = sum(1 for ref in self.repo.refs)
        if ref_count > 0:
            # 判断这个仓库是否有过提交
            try:
                shell(f"git submodule foreach git checkout .")
                shell(f"git submodule foreach git clean -f")
            except Exception as e:
                logger.warning('clean submodules failed')

            # 初始化一下子项目
            shell(f"git submodule update --init --recursive --progress")
            self.repo.iter_submodules()
            sms = self.repo.submodules
            if sms and len(sms) > 0:
                is_init = True

        if not is_init:
            logger.error("Not initialized yet, please execute the [trident init] command first")
            raise Exception("Not initialized yet, please execute the [trident init] command first")

        sync_task_map = config.sync

        try:
            for key in sync_task_map:
                sync_task: SyncTask = sync_task_map[key]
                # 执行同步任务
                task_executor = TaskExecutor(self.work_root, self.config, self.status, sms, self.http, sync_task)
                task_executor.do_task()

            self.config.status.success = True

            # 所有任务已完成
            # 提交同步仓库的变更
            self.commit_work_repo()
            self.repo.close()
        finally:
            self.render_result(sync_task_map)
            logger.info(text_center("sync end"))

    def render_result(self, conf_sync_map):
        def right(target: str, res: bool, label_length=8):
            return f"{target.rjust(label_length, ' ')}:{'✅' if res else '🚫'}"

        def fill(string: str):
            return string.ljust(15, " ")

        cs = self.config.status
        result = text_center(right('result', cs.success, 0))
        for key in conf_sync_map:
            t: SyncTask = conf_sync_map[key]
            s = t.status
            task_result = f"\n 🏹 {fill(t.key)} --> {right('success', s.success)} {right('copy', s.copy)} {right('change', s.change)} {right('commit', s.commit)} {right('push', s.push)} {right('pr', s.pr)} {right('merge', s.merge)}"
            result += task_result
        result += f"\n 🔱 {fill('sync_work_repo')} --> {right('change', cs.change)} {right('commit', cs.commit)} {right('push', cs.push)} "
        # 输出结果
        logger.info(result)

    def commit_work_repo(self):
        now = datetime.datetime.now()
        message = f"🔱: sync all task at {now} [trident-sync]"
        os.chdir(self.work_root)
        save_work_repo(self.repo, message, self.config.options.push, status=self.config.status)


class TaskExecutor:
    def __init__(self, work_root, config: Config, status: dict, sms, http, sync_task: SyncTask):
        self.key = sync_task.key
        self.work_root = work_root
        self.sync_task = sync_task
        self.sms = sms
        self.task_src = sync_task.src
        self.task_target = sync_task.target

        self.conf_options = config.options

        self.status = status
        self.http = http

        self.conf_src_repo = self.task_src.repo_ref
        self.conf_target_repo = self.task_target.repo_ref
        self.repo_src = sms[self.task_src.repo].module()
        self.repo_target = sms[self.task_target.repo].module()

    def do_task(self):

        logger.info(text_center(f"【{self.key}】 task start"))
        time.sleep(0.2)

        # 同步任务开始
        # 更新源仓库代码
        self.pull_src_repo()
        # 当前目录切换到目标项目
        os.chdir(self.repo_target.working_dir)
        # 先强制切换回主分支
        force_checkout_main_branch(self.task_target.repo_ref)
        # 创建同步分支，并checkout
        is_first = checkout_branch(self.repo_target, self.task_target.branch)
        # 开始复制文件
        os.chdir(self.work_root)
        self.do_sync(is_first)
        os.chdir(self.repo_target.working_dir)
        # 提交代码
        self.do_commit()
        # push更新
        has_push = self.do_push()
        # 创建PR
        self.do_pull_request(has_push)
        # 切换回主分支
        force_checkout_main_branch(self.task_target.repo_ref)

        logger.info(text_center(f"【{self.key}】 task complete"))
        self.sync_task.status.success = True
        self.repo_src.close()
        self.repo_target.close()

    def pull_src_repo(self):
        logger.info(f"update src repo :{self.task_src.repo_ref.url}")
        shell(f"cd {self.repo_src.working_dir} && git checkout {self.task_src.repo_ref.branch} -f && git pull")
        logger.info(f"update submodule of src repo")
        shell(f"cd {self.repo_src.working_dir} && git submodule update --init --recursive --progress ")
        logger.info(f"update src repo success")

    def do_sync(self, is_first):
        dir_src_sync = f"{self.repo_src.working_dir}/{self.task_src.dir}"
        dir_target_sync = f"{self.repo_target.working_dir}/{self.task_target.dir}"
        logger.info(f"sync dir：{dir_src_sync}->{dir_target_sync}")
        # 检查源仓库目录是否有文件，如果没有文件，可能初始化仓库不正常
        src_is_blank = is_blank_dir(dir_src_sync)
        if src_is_blank:
            raise Exception(
                f"The src repo dir <{dir_src_sync}> is empty. It may not be fully initialized. Try to enter this directory and execute the [git pull] command")

        remove_target_dir = True
        if is_first:
            # 第一次同步，目标目录必须为空
            target_is_blank = is_blank_dir(dir_target_sync)
            if not target_is_blank:
                logger.warning(
                    f"For the first time, the target repo dir <{dir_target_sync}> is not empty")
                logger.warning(
                    f"Please make sure that the dir is a copy of a version of the src repo, otherwise please change the directory!!")
                logger.warning(
                    f"If you are sure that the directory is a copy of the source repository, you can try configuring \
                    <sync.[task].target.allow_reset_to_root:true> and reruning [trident sync] command ,This will \
                    reset the sync_branch to first commit to see if an earlier version had the \
                    directory.")
                if not self.task_target.allow_reset_to_root:
                    raise Exception(
                        f"the target repo dir <{dir_target_sync}> is not empty, and allow_reset_to_root is False")
                else:
                    logger.info(f"The allow_reset_to_root is True, Trying to reset the sync_branch to root commit")
                    root_hash = shell("git rev-list --max-parents=0 HEAD", get_out=True)
                    shell(f"git reset {root_hash.strip()}")
                    shell("git clean -df ")
                    logger.info(f"Reset the sync_branch to root commit success")
                    # 再次检测目录是否为空
                    target_is_blank = is_blank_dir(dir_target_sync)
                    if not target_is_blank:
                        logger.warning(
                            f"The target repository directory <{dir_target_sync}> is still not empty, Some changes may be lost!!!")
                        # 此时不移除目标目录
                        remove_target_dir = False

        if remove_target_dir and os.path.exists(dir_target_sync):
            logger.info(f"remove <{dir_target_sync}> ...")
            shutil.rmtree(dir_target_sync)
            time.sleep(0.2)
        logger.info(f"copy files")
        sync_func(self.sync_task, dir_src_sync, dir_target_sync)
        git_file = f"{dir_target_sync}/.git"
        if os.path.exists(git_file):
            os.unlink(git_file)
        logger.info(f"【{self.key}】 Copy completed, ready to submit : {self.task_target.dir}")
        time.sleep(1)
        self.sync_task.status.copy = True

    def do_commit(self):
        shell(f"git add .")
        time.sleep(1)
        count = get_git_modify_file_count()
        time.sleep(1)
        logger.info(f"modify count : {count}")
        key = self.key
        if count <= 0:
            logger.info(f"【{key}】 No change, no need to submit")
            return False
        else:
            self.sync_task.status.change = True
            last_commit = get_dict_value(self.status, f"sync.{key}.last_commit_src")
            messsges = collection_commit_message(self.repo_src, self.task_src.repo_ref.branch, last_commit)
            body = ""
            for msg in messsges:
                body += msg + "\n"
            now = datetime.datetime.now()
            message = f"🔱: [{key}] sync upgrade with {len(messsges)} commits [trident-sync] "
            # 提交更新
            shell(f'git commit -m "{message}" -m "{body}"')
            # repo_target.index.commit(f"sync {key} success [{now}]")
            logger.info(f"【{key}】 submit success")
            time.sleep(0.2)
            # 记录最后提交hash
            src_last_hash = self.repo_src.head.commit.hexsha
            target_last_hash = self.repo_target.head.commit.hexsha

            set_dict_value(self.status, f"sync.{key}.last_commit_src", src_last_hash)
            set_dict_value(self.status, f"sync.{key}.last_commit_target", target_last_hash)
            save_status(self.work_root, self.status)
            self.sync_task.status.commit = True
            return True

    def do_push(self):
        if not self.conf_options.push:
            return False
        logger.info("Check if push is needed")
        # 检测是否需要push
        key = self.key
        need_push = check_need_push(self.repo_target, self.task_target.branch)
        if need_push is False:
            logger.info("No commit to push")
            return False
        else:
            logger.info("need push")
            logger.info(f"【{key}】 pushing")
            shell(f'git push --set-upstream origin {self.task_target.branch}')
            logger.info(f"【{key}】 push success")
            time.sleep(0.2)
            self.sync_task.status.push = True
            return True

    def do_pull_request(self, has_push):
        key = self.key
        if not self.conf_options.pull_request:
            return False
        if not has_push:
            return False
        token = self.task_target.repo_ref.token
        repo_type = self.task_target.repo_ref.type
        auto_merge = self.conf_target_repo.auto_merge
        if not repo_type:
            logger.warning(f"[repo:{self.task_target.repo}] type is not configured, Unable to create PR")
            return False
        if not token:
            logger.warning(f"[repo:{self.task_target.repo}] token is not configured, Unable to create PR")
            return False
        else:
            client = api_clients[repo_type](self.http, token, self.task_target.repo_ref.url)
            title = f"[{key}] sync upgrade [trident-sync]"
            body = f"{self.task_src.repo}:{self.conf_src_repo.branch}:{self.task_src.dir} -> {self.task_target.repo}:\
                {self.conf_target_repo.branch}:{self.task_target.dir} "
            logger.info(
                f"Ready to create PR, {self.task_target.branch} -> {self.conf_target_repo.branch} , url:{self.conf_target_repo.url}")
            try:
                pull_id, merged = client.create_pull_request(title, body, self.task_target.branch,
                                                             self.conf_target_repo.branch,
                                                             auto_merge=auto_merge)
                self.sync_task.status.pr = True
                if merged:
                    self.sync_task.status.merge = True
            except Exception as e:
                # logger.opt(exception=e).error("提交PR出错")
                logger.error(f"Error creating PR：{e}")
            time.sleep(0.2)
            return True



import os

from lib.model.opts import Options
from lib.model.repo import RepoConf
from lib.model.sync import SyncTask
from lib.util import get_dict_value


class RunStatus:
    success: bool = False
    change: bool = False
    commit: bool = False
    push: bool = False


class Config:
    repo = {}
    sync = {}
    options: Options

    status: RunStatus

    def __init__(self, conf_dict):
        conf_repo = get_dict_value(conf_dict, 'repo')
        for key in conf_repo:
            self.repo[key] = RepoConf(key, conf_repo[key])
        conf_sync = get_dict_value(conf_dict, 'sync')
        for key in conf_sync:
            self.sync[key] = SyncTask(key, conf_sync[key], self.repo)
        conf_options = get_dict_value(conf_dict, 'options')
        self.status = RunStatus()
        self.options = Options(conf_options)

    def set_default_token(self, token=None):
        for key in self.repo:
            repo: RepoConf = self.repo[key]
            if repo.token:
                continue
            if not token:
                token = os.getenv(f'{repo.type}_token')
                if not token:
                    token = os.getenv(f'trident_token_{repo.type}')
            if token:
                repo.token = token



'''
repo:
  test: # 你的项目（接受同步项目），可以任意命名
    url: "https://github.com/handsfree-work/trident-test"           # 仓库地址
    path: "target"                    # submodule子路径
    branch: "main"                    # 你的代码开发主分支（接受合并的分支）例如dev、main等
    token: ""                         # 仓库token，用于提交PR, 前往 https://github.com/settings/tokens 创建token
    type: github                      # 仓库类型，用于提交PR，可选项：[github/gitee/gitea/gitlab]
'''
from lib.util import merge_from_dict


class RepoConf:
    key: str = None

    # submodule相关
    url: str = None
    branch: str = None
    path: str = None

    # pr 相关
    token: str = None
    type: str = None
    auto_merge: bool = True

    def __init__(self, key, conf_dict: dict):
        self.key = key
        merge_from_dict(self, conf_dict)
        if not self.url or not self.branch or not self.path:
            raise Exception(f"repo<{key}> 中 url/branch/path 必须配置")

        if self.url.startswith("https://github.com"):
            self.type = 'github'
        elif self.url.startswith("https://gitee.com"):
            self.type = 'gitee'



from pathlib import Path

from lib.util import read_file


def project_root() -> Path:
    """Find the project root directory by locating pyproject.toml."""
    current_file = Path(__file__)
    for parent_directory in current_file.parents:
        if (parent_directory / "pyproject.toml").is_file():
            return parent_directory
    raise FileNotFoundError("Could not find project root containing pyproject.toml")


def get_version():
    try:
        # Probably this is the pyproject.toml of a development install
        path_to_pyproject_toml = project_root() / "pyproject.toml"
    except FileNotFoundError:
        # Probably not a development install
        path_to_pyproject_toml = None

    if path_to_pyproject_toml is not None:
        content = read_file(path_to_pyproject_toml)
        lines = content.split('\n')
        for line in lines:
            line = line.strip()
            if line.startswith("version = "):
                version = line.replace("version = ", "").replace("\"", "")
                return version
    return "0.0.0"



import datetime
import os
import time

from git import Repo

from lib.logger import logger
from lib.model.config import RunStatus
from lib.util import shell, check_need_push
from lib.util_git import get_git_modify_file_count


def save_work_repo(repo: Repo, message, push=True, status: RunStatus = None):
    shell("git add .")
    if status is None:
        status = RunStatus()
    count = get_git_modify_file_count()

    if 'origin' in repo.remotes:
        try:
            shell("git pull")
        except Exception as e:
            logger.warning('git pull failed')

    if count <= 0:
        logger.info("No modification, no need to submit")
    else:
        status.change = True
        time.sleep(1)
        shell(f'git commit -m "{message}"')
        status.commit = True

    if push:
        need_push = check_need_push(repo, repo.active_branch)
        if need_push is None:
            logger.warning(
                "Skip push，The remote address is not set for the current repository")
            logger.warning(
                "Use the [trident remote --url=<repo_url>] command to set the remote address of the repository and save the synchronization progress")

        elif need_push is True:
            shell(f"git push")
            status.push = True
        else:
            logger.info("No need to push")



import os
import shutil
import stat
import subprocess as sp
import re
from lib.logger import logger


def re_pick(re_str, input_str, flags=0):
    # inputStr = "['{0: 203, 11: 1627438682 [2021-07-28 10:18:02],  12:36 [通过蓝牙更改错误密码锁定计数], 13: 770449129, 19: 3, 20: 0, 100: 2141634486}']"

    reg = re.compile(re_str, flags)  # 增加匹配效率的 S 多行匹配
    lists = re.findall(reg, str(input_str))
    if len(lists) > 0:
        return lists[0]
    return []


def get_dict_value(dict_target, key, def_value=None):
    arr = str.split(key, '.')
    value = None
    parent = dict_target
    for key in arr:
        if parent and key and key in parent:
            value = parent[key]
            parent = value
        else:
            value = None
    if value is None:
        value = def_value
    return value


def set_dict_value(dict_target: object, key: str, value: object) -> object:
    arr = str.split(key, '.')
    parent = dict_target
    last_key = arr[len(arr) - 1]
    for key in arr:
        if last_key == key:
            parent[key] = value
            return
        if key not in parent:
            parent[key] = {}
        parent = parent[key]


def shell(cmd, ignore_errors=False, get_out=False):
    logger.info(cmd)
    out = None
    if get_out:
        out = sp.PIPE

    p = sp.run(cmd, shell=True, encoding='utf-8', stdout=out)
    if p.returncode != 0 and not ignore_errors:
        logger.debug(p.stdout)
        logger.debug(p.stderr)
        raise Exception(p.stderr)
    if get_out and p.stdout:
        logger.debug(p.stdout)
    return p.stdout


def read_file(file_path):
    if not os.path.exists(file_path):
        return None
    fo = open(file_path, "r", encoding='utf-8')
    content = fo.read()
    fo.close()
    return content


def save_file(file_path, content):
    dir_path = os.path.dirname(file_path)
    if not os.path.exists(dir_path):
        os.makedirs(dir_path)
    fo = open(file_path, "w", encoding='utf-8')
    fo.write(content)
    fo.close()


def check_need_push(repo, branch):
    """
    检查是否需要push，hash相等返回false，hash不相等返回true，没有远程分支返回None
    :param repo:
    :param branch:
    :return:
    """
    local_hash = repo.head.commit.hexsha
    remote_hash = None
    refs = repo.refs
    logger.info(f"refs:{refs}")
    origin_key = f"origin/{branch}"
    if origin_key in refs:
        remote_hash = refs[origin_key].commit.hexsha
    else:
        return None

    logger.info(f"local_hash:{local_hash} -> remote_hash:{remote_hash} ")
    if local_hash == remote_hash:
        return False
    return True


def merge_from_dict(obj, dic):
    for key in dic:
        setattr(obj, key, dic[key])


def rm_dir(root):
    def readonly_handler(func, path, exe_info):
        os.chmod(path, stat.S_IWRITE)
        func(path)

    shutil.rmtree(root, onerror=readonly_handler)


def get_arg(args, key):
    value = args[key]
    if isinstance(value, list):
        if len(value) > 0:
            value = value[0]
        else:
            return None
    return value


def is_blank_dir(root):
    """
    检查目录是否为空，如果目录不存在也返回True
    """
    if not os.path.exists(root):
        return True
    is_blank = True
    for file in os.listdir(root):
        if file == '.git':
            continue
        is_blank = False
        break
    return is_blank



import os
import time

import git

from lib.handler.helper import save_work_repo
from lib.logger import logger
from lib.model.config import Config
from lib.model.repo import RepoConf
from lib.util import shell, save_file


class InitHandler:

    def __init__(self, work_root, config):
        self.work_root = work_root
        self.config = config

    def handle(self):
        """
        处理 init 命令
        """
        work_root = self.work_root
        config: Config = self.config
        logger.info(f"git init : {work_root}")
        os.chdir(work_root)
        shell('git init')
        repo = git.Repo(path=work_root)
        if len(repo.heads) == 0:
            self.save_ignore_file()
            shell("git add .")
            time.sleep(1)
            shell('git commit -m "🔱: sync init start [trident-sync]"')
        logger.info("get submodules")
        sms = repo.iter_submodules()
        conf_repos = config.repo
        conf_options = config.options
        conf_repo_root = conf_options.repo_root
        for key in conf_repos:
            added = False
            for module in sms:
                if key == module.name:
                    logger.info(f"{key} has been added to the submodule")
                    added = True
                    break
            if added:
                continue
            item: RepoConf = conf_repos[key]
            logger.info(f"add submodule:{item.url}")
            path = f"{conf_repo_root}/{item.path}"
            # repo.create_submodule(key, path, url=item['url'], branch=item['branch'])
            shell(f"git submodule add -b {item.branch} --name {key} {item.url} {path}")

        logger.info("Update all submodule")

        shell(f"git submodule update --init --recursive --progress")
        repo.iter_submodules()
        save_work_repo(repo, '🔱: sync init [trident-sync]', push=config.options.push)
        os.chdir(os.getcwd())
        logger.info("init success")
        repo.close()

    def save_ignore_file(self):
        ignore_file = f"{self.work_root}/.gitignore"
        ignore = '''
.idea
.vscode
.git
__pycache__
        '''
        save_file(ignore_file, ignore)



"""
异构仓库同步升级工具

Precondition:
    1. install git
Usage:
    trident init [-r ROOT] [-c CONFIG]
    trident sync [-r ROOT] [-c CONFIG] [-t TOKEN] [-i]
    trident remote [-r ROOT] [-u URL] [-f]
    trident version
    trident -h
Options:
    -h,--help           show help menu, 显示帮助菜单
    -c,--config=CONFIG  config file path, 配置文件  [default: sync.yaml]
    -r,--root=ROOT      root dir, 根目录  [default: .]
    -t,--token=TOKEN    PR token
    -u,--url=URL        remote git url, 远程地址
    -f,--force          force push, 强制推送
    -i,--init           init first, 同步前先初始化
Example:
    trident init
    trident sync
    trident remote --url=https://github.com/handsfree-work/trident-test-sync
"""
import os

import yaml
from docopt import docopt

from lib.handler.init import InitHandler
from lib.handler.remote import RemoteHandler
from lib.handler.sync import SyncHandler
from lib.logger import logger
from lib.model.config import Config
from lib.util import get_arg
from lib.version import get_version


def cli():
    """
    异构仓库同步升级工具入口
    """
    args = docopt(__doc__)

    version = get_version()

    print(f'''
 
                  ████████╗██████╗ ██╗██████╗ ███████╗███╗   ██╗████████╗  
                  ╚══██╔══╝██╔══██╗██║██╔══██╗██╔════╝████╗  ██║╚══██╔══╝  
                     ██║   ██████╔╝██║██║  ██║█████╗  ██╔██╗ ██║   ██║     
                     ██║   ██╔══██╗██║██║  ██║██╔══╝  ██║╚██╗██║   ██║     
                     ██║   ██║  ██║██║██████╔╝███████╗██║ ╚████║   ██║     
                     ╚═╝   ╚═╝  ╚═╝╚═╝╚═════╝ ╚══════╝╚═╝  ╚═══╝   ╚═╝     
                      https://github.com/handsfree-work/trident-sync       
                    Don't be stingy with your star ( 请不要吝啬你的star )
                       Copyright © 2023 greper@handsfree.work v{version}    
                                                  
        ''')
    root = get_root(args)
    if not os.path.exists(root):
        os.makedirs(root)

    arg_config = get_arg(args, '--config')
    config_dict = read_config(arg_config)

    os.chdir(root)
    if args['remote']:
        remote_url = args['--url']
        force = args['--force']
        RemoteHandler(root, remote_url=remote_url, force=force).handle()
        return

    token_from_args = get_arg(args, '--token')
    config = Config(config_dict)
    config.set_default_token(token_from_args)

    if args['init']:
        InitHandler(root, config).handle()
    elif args['sync']:
        init_first = get_arg(args, '--init')
        if init_first:
            logger.info("init first")
            InitHandler(root, config).handle()
        SyncHandler(root, config).handle()
    else:
        logger.info(__doc__)


def read_config(arg_config='./sync.yaml'):
    config_file = os.path.abspath(f"{arg_config}")
    f = open(config_file, 'r', encoding='utf-8')
    return yaml.load(f, Loader=yaml.FullLoader)


def get_root(args):
    root = get_arg(args, '--root')
    return os.path.abspath(root)


if __name__ == '__main__':
    cli()



import time

from lib.api.abstract_client import AbstractClient
from lib.logger import logger
from lib.util import get_dict_value, re_pick


class GiteeClient(AbstractClient):
    '''gitee api'''
    '''
    https://gitee.com/api/v5/swagger#/postV5ReposOwnerRepoPulls
    '''

    def __init__(self, http, token, url):
        super().__init__(http, token, url)
        self.api_root = "https://gitee.com/api/v5"

    def post_pull_request(self, title, body, head_branch, base_branch):
        api = f"{self.api_root}/repos/{self.repo_path}/pulls"
        res = self.http.post(api, data={
            "access_token": self.token,
            "title": title,
            "body": body,
            "head": f"{self.owner}:{head_branch}",
            "base": base_branch
        }, res_is_standard=False, res_is_json=True)
        return res

    def query_pull_request(self, head_branch, base_branch, state='open'):
        api = f"{self.api_root}/repos/{self.repo_path}/pulls?access_token={self.token}&head={self.owner}:{head_branch}&base={base_branch}&state={state}"
        res = self.http.get(api, res_is_standard=False, res_is_json=True)
        if len(res) > 0:
            return res[0]
        return None

    def get_pull_request_detail(self, pull_id):
        api = f"{self.api_root}/repos/{self.repo_path}/pulls/{pull_id}?access_token={self.token}"
        res = self.http.get(api, res_is_standard=False, res_is_json=True)
        return res

    def post_merge(self, pull_id, detail):
        if detail['assignees_number'] > 0:
            self.post_review(pull_id)
        if detail['testers_number'] > 0:
            self.post_test(pull_id)

        api = f"{self.api_root}/repos/{self.repo_path}/pulls/{pull_id}/merge"
        self.http.put(api, data={
            "access_token": self.token
        }, res_is_standard=False, res_is_json=True)

    def post_review(self, pull_id):
        logger.info("force approve review")
        api = f"{self.api_root}/repos/{self.repo_path}/pulls/{pull_id}/review"
        self.http.post(api, data={
            "access_token": self.token,
            "force": True
        }, res_is_standard=False, res_is_json=False)

    def post_test(self, pull_id):
        logger.info("force approve test")
        api = f"{self.api_root}/repos/{self.repo_path}/pulls/{pull_id}/test"
        self.http.post(api, data={
            "access_token": self.token,
            "force": True
        }, res_is_standard=False, res_is_json=False)



'''
sync:
    task1:
        src: # 源仓库
          repo: fs-admin                  # 源仓库名称，上面repo配置的仓库名称引用
          dir: '.'                        #要同步给target的目录
        target: #目标仓库
          repo: test                      # 目标仓库名称，上面repo配置的仓库名称引用
          dir: 'package/ui/certd-client'  # 接收src同步过来的目录
          branch: 'client_sync'           # 同步分支名称（需要配置一个未被占用的分支名称）
'''
from lib.model.repo import RepoConf
from lib.util import merge_from_dict


class SyncTaskSrc:
    repo: str = None
    dir: str = None
    repo_ref: RepoConf

    def __init__(self, conf_dict):
        merge_from_dict(self, conf_dict)
        if not self.repo or not self.dir:
            raise Exception("sync.[key].src 中 < repo/dir > 必须配置")


class SyncTaskTarget:
    repo: str = None
    dir: str = None
    branch: str = None
    allow_reset_to_root: bool = True
    remove_dir_before_copy: bool = True
    repo_ref: RepoConf

    def __init__(self, conf_dict):
        merge_from_dict(self, conf_dict)
        if not self.repo or not self.dir or not self.branch:
            raise Exception("sync.[key].target 中 < repo/dir/branch > 必须配置")


class SyncStatus:
    success: bool = False
    change: bool = False
    copy: bool = False
    commit: bool = False
    push: bool = False
    pr: bool = False
    merge: bool = False


class SyncTask:
    key: str = None
    src: SyncTaskSrc = None
    target: SyncTaskTarget = None
    copy_script: str = None

    status: SyncStatus

    def __init__(self, key, conf_sync: dict, repo_map):
        self.key = key

        if 'src' not in conf_sync:
            raise Exception(f"sync.{key}.src 必须配置")
        if 'target' not in conf_sync:
            raise Exception(f"sync.{key}.target 必须配置")

        merge_from_dict(self, conf_sync)
        self.status = SyncStatus()

        self.src = SyncTaskSrc(conf_sync['src'])
        self.target = SyncTaskTarget(conf_sync['target'])

        self.set_repo_ref(self.src, repo_map)
        self.set_repo_ref(self.target, repo_map)

    def set_repo_ref(self, task, repo_list):
        if task.repo in repo_list:
            task.repo_ref = repo_list[task.repo]
        else:
            raise Exception(f"任务[{self.key}]的{self.src.repo} 仓库配置不存在，请检查repo配置")



import json

import requests
import urllib

from lib.logger import logger


class HttpException(Exception):
    # 自定义异常类型的初始化
    def __init__(self, msg, code, data):
        self.msg = msg
        self.code = code
        self.data = data

    # 返回异常类对象的说明信息
    def __str__(self):
        return self.msg


def standard_res_handle(res):
    if res['code'] == 0:
        return res['data']
    raise Exception(res['msg'], res['code'])


class Http:
    def __init__(self, use_system_proxy=True, proxy_fix=True):
        self.proxies = {'http': None, 'https': None}
        if use_system_proxy:
            proxies = urllib.request.getproxies()
            if proxy_fix and "https" in proxies:
                https_proxy = proxies['https']
                if https_proxy.startswith('https://'):
                    proxies['https'] = https_proxy.replace("https://", "http://")
            self.proxies = proxies
        self.verify = True
        self.logger = logger

    def set_proxies(self, proxies):
        self.proxies = proxies

    def options(self, url, cookies=None, headers=None, **kwargs):
        self.logger.debug(f"http request[options] url:{url}")
        if headers is None:
            headers = {}
        if 'Content-Type' not in headers:
            headers['Content-Type'] = 'application/json'
        session = requests.Session()
        session.trust_env = False
        response = session.options(url, proxies=self.proxies, headers=headers, cookies=cookies, **kwargs)
        self.logger.debug("http response: " + response.text)
        return

    def post(self, url, data, res_is_json=True, res_is_standard=True, cookies=None, headers=None, **kwargs):
        self.logger.debug(f"http request[post] url:{url}")
        if headers is None:
            headers = {}
        if 'Content-Type' not in headers:
            headers['Content-Type'] = 'application/json'
        session = requests.Session()
        session.headers.clear()
        session.headers.update(headers)
        session.trust_env = False
        response = session.post(url, json=data, proxies=self.proxies, headers=headers, cookies=cookies,
                                verify=self.verify, **kwargs)
        self.logger.debug("http response: " + response.text)
        return self.res_handle(response, res_is_json, res_is_standard)

    def put(self, url, data, res_is_json=True, res_is_standard=True, cookies=None, headers=None, **kwargs):
        self.logger.debug(f"http request[put] url:{url}")
        if headers is None:
            headers = {}
        if 'Content-Type' not in headers:
            headers['Content-Type'] = 'application/json'
        session = requests.Session()
        session.headers.clear()
        session.headers.update(headers)
        session.trust_env = False
        response = session.put(url, json=data, proxies=self.proxies, headers=headers, cookies=cookies,
                               verify=self.verify, **kwargs)
        self.logger.debug("http response: " + response.text)
        return self.res_handle(response, res_is_json, res_is_standard)

    def get(self, url, headers=None, res_is_json=True, res_is_standard=True):
        self.logger.debug(f"http request[get] url:{url}")
        if headers is None:
            headers = {}
        if 'Content-Type' not in headers:
            headers['Content-Type'] = 'application/json'
        session = requests.Session()
        session.trust_env = False
        response = session.get(url, proxies=self.proxies, verify=self.verify, headers=headers)
        self.logger.debug("http response: " + response.text)
        return self.res_handle(response, res_is_json, res_is_standard)

    def download(self, url, save_path, on_progress=None):
        session = requests.Session()
        session.trust_env = False
        down_res = session.get(url=url, proxies=self.proxies, stream=True)
        chunk_size = 1024 * 1024
        downloaded = 0
        content_size = int(down_res.headers['content-length'])  # 内容体总大小
        with open(save_path, "wb") as file:
            for data in down_res.iter_content(chunk_size=chunk_size):
                file.write(data)
                downloaded += len(data)
                if on_progress is not None:
                    on_progress({
                        'total': content_size,
                        'downloaded': downloaded,
                        'status': 'downloading'
                    })
        if on_progress:
            on_progress({
                'total': content_size,
                'downloaded': downloaded,
                'status': 'finished'
            })

    def res_handle(self, response, res_is_json, res_is_standard):

        if response.status_code < 200 or response.status_code > 299:
            return self.error_handle(response)

        if res_is_json:
            res = response.json()
            if res_is_standard:
                return standard_res_handle(res)
            return res
        else:
            return response.text

    def error_handle(self, response):
        data = {}
        try:
            data = json.loads(response.text)
        except Exception as e:
            pass
        raise HttpException(f'请求错误：({response.status_code}):{response.text}', 1, data)



from lib.api.abstract_client import AbstractClient


class GithubClient(AbstractClient):
    '''
            https://docs.github.com/en/rest/pulls/pulls?apiVersion=2022-11-28

            curl \
              -X POST \
              -H "Accept: application/vnd.github+json" \
              -H "Authorization: Bearer <YOUR-TOKEN>"\
              -H "X-GitHub-Api-Version: 2022-11-28" \
              https://api.github.com/repos/OWNER/REPO/pulls \
              -d '{"title":"Amazing new feature","body":"Please pull these awesome changes in!","head":"octocat:new-feature","base":"master"}'
            '''

    '''github api'''
    # 下面定义了一个类属性
    token = 'token'
    http = None
    url = None
    repo_path = None
    owner = None
    repo = None
    headers = None;

    def __init__(self, http, token, url):
        super().__init__(http, token, url)

        self.headers = {
            "Accept": "application/vnd.github+json",
            "Authorization": f"Bearer {self.token}",
            "X-GitHub-Api-Version": "2022-11-28"
        }

    def post_pull_request(self, title, body, head_branch, base_branch):
        """
        head_branch 来源分支
        base_branch 被合并分支
        """
        api = f"https://api.github.com/repos/{self.repo_path}/pulls"
        res = self.http.post(api, data={
            "title": title,
            "body": body,
            "head": f"{self.owner}:{head_branch}",
            "base": base_branch,
            "maintainer_can_modify": True
        }, headers=self.headers, res_is_standard=False, res_is_json=True)
        return res

    def query_pull_request(self, head_branch, base_branch, state='open'):
        api = f"https://api.github.com/repos/{self.repo_path}/pulls?head={self.owner}:{head_branch}&base={base_branch}&state={state}"
        res = self.http.get(api, headers=self.headers, res_is_standard=False, res_is_json=True)
        if len(res) > 0:
            return res[0]
        return None

    def get_pull_request_detail(self, pull_id):
        api = f"https://api.github.com/repos/{self.repo_path}/pulls/{pull_id}"
        res = self.http.get(api, headers=self.headers, res_is_standard=False, res_is_json=True)
        return res

    def post_merge(self, pull_id, detail):
        api = f"https://api.github.com/repos/{self.repo_path}/pulls/{pull_id}/merge"
        self.http.put(api, data={}, headers=self.headers, res_is_standard=False, res_is_json=True)



import os
import time

from git import Repo

from lib.logger import logger
from lib.model.config import RunStatus
from lib.model.repo import RepoConf
from lib.util import shell, check_need_push



def add_and_commit(message):
    shell("git add .")
    count = get_git_modify_file_count()
    if count > 0:
        time.sleep(1)
        shell(f'git commit -m "{message}"')


def get_git_modify_file_count():
    ret = shell(f"git status", get_out=True)
    lines = ret.split("\n")
    file_list = []
    # 忽略的package列表

    count = 0
    for line in lines:
        start = line.find(':   ')
        if start < 0:
            continue
        start += 1
        file = line[start:].strip()
        count += 1
    return count


def force_checkout_main_branch(conf_repo_ref: RepoConf):
    # 切换回主分支
    shell(f"git checkout -f {conf_repo_ref.branch}")
    time.sleep(1)


def checkout_branch(repo: Repo, branch: str):
    '''
    创建或更新分支，如果远程有，则从远程拉取
    '''
    # 看看远程是否有对应分支
    logger.info(f"checkout branch：{branch}")

    origin_key = f"origin/{branch}"
    origin_exists = False
    local_exists = False
    if origin_key in repo.refs:
        origin_exists = True

    if branch in repo.heads:
        local_exists = True

    is_first = False
    if origin_exists and not local_exists:
        # 远程有，本地没有，从远程拉取
        shell(f"git branch {branch} --track origin/{branch}")
    elif not origin_exists and not local_exists:
        # 两边都没有，本地创建
        shell(f"git branch {branch}")
        is_first = True
    elif origin_exists and local_exists:
        # 两边都有
        shell(f"git checkout {branch}")
        shell(f"git pull")
    time.sleep(1)
    shell(f"git checkout {branch}")
    time.sleep(1)
    return is_first


def collection_commit_message(repo, branch, last_commit=None, max_count=20):
    # 准备commit文本
    commits = repo.iter_commits(branch, max_count=max_count)
    messages = []
    more = "..."
    for item in commits:
        if item.hexsha == last_commit:
            more = ""
            break
        messages.append(item.message.strip())
    messages.append(more)
    return messages



import os

import git

from lib.logger import logger
from lib.util import shell


class RemoteHandler:

    def __init__(self, work_root, remote_url=None, force=False):
        self.work_root = work_root
        self.remote_url = remote_url
        self.force = force

    def handle(self):
        os.chdir(self.work_root)
        repo = git.Repo(path=self.work_root)
        cur_branch_name = repo.head.reference
        url = self.remote_url
        if 'origin' in repo.remotes:
            logger.info("The remote origin already exists and no url parameter is needed")
            shell("git pull")
        else:
            if not url:
                logger.info(
                    "Please use the [trident remote --url=<sync_work_repo_git_url>] command to set the remote address first")
                return
            else:
                shell(f"git remote add origin {url}")
                # origin = repo.create_remote("origin", url)
                logger.info('add remote origin success:' + url)
        force = ""
        if self.force:
            force = " -f "
        shell(f"git push -u {force} origin {cur_branch_name}")
        logger.info('push success')



import time
from abc import abstractmethod

from lib.logger import logger
from lib.util import re_pick


def pick_from_url(git_url):
    re_str = "(http[s]?)://([^/]+)/([^/]+)/([^/\\.]+)"
    res = re_pick(re_str, git_url)
    if len(res) < 4:
        raise Exception(f"异常的git_url: {git_url}")
    return {
        'protocol': res[0],
        'host': res[1],
        'owner': res[2],
        'repo': res[3],
    }


class AbstractClient:
    '''repo api'''
    token = 'token'
    http = None
    url = None
    repo_path = None
    owner = None
    repo = None
    api_root = None

    def __init__(self, http, token, url):
        self.token = token
        self.http = http
        self.url = url
        res = pick_from_url(url)
        self.owner = res['owner']
        self.repo = res['repo']
        self.url_prefix = f"{res['protocol']}://{res['host']}"
        self.repo_path = f"{self.owner}/{self.repo}"

    def create_pull_request(self, title, body, head_branch, base_branch, auto_merge=True):
        '''
        https://try.gitea.io/api/swagger#/repository/repoCreatePullRequest

         curl -X 'POST' \
          'http://gitea.baode.docmirror.cn/api/v1/repos/{owner}/{repo}/pulls?access_token=1a5461e976c423068cc7677e608681b3d992eefe' \
          -H 'accept: application/json'
        '''

        pull_res = self.query_pull_request(head_branch, base_branch)
        if pull_res is None:
            pull_res = self.post_pull_request(title, body, head_branch, base_branch)
        pull_id = pull_res['number']
        merged = False
        if auto_merge:
            merged = self.auto_merge(pull_id)
        return pull_id, merged

    def can_auto_merge(self, res):
        if 'mergeable' not in res:
            logger.warning("Status unknown, please manually merge PR")
            return None
        mergeable = res['mergeable']
        if mergeable is True:
            return True
        if mergeable is False:
            logger.warning("There may be conflicts, please merge PR manually")
            return False
        logger.warning("Status unknown, please manually merge PR")
        return None

    def auto_merge(self, pull_id):
        logger.info("Check for can auto merge after 5 seconds")
        time.sleep(5)
        res = self.get_pull_request_detail(pull_id)
        can_merge = self.can_auto_merge(res)
        if can_merge is True:
            logger.info("PR will be auto merged")
            # 准备自动合并
            self.post_merge(pull_id, res)
            logger.info("Auto merge success")
            return True
        return False

    @abstractmethod
    def post_pull_request(self, title, body, head_branch, base_branch):
        pass

    @abstractmethod
    def query_pull_request(self, head_branch, base_branch):
        pass

    @abstractmethod
    def get_pull_request_detail(self, pull_id):
        pass

    @abstractmethod
    def post_merge(self, pull_id, detail):
        pass



from lib.api.gitea import GiteaClient
from lib.api.gitee import GiteeClient
from lib.api.github import GithubClient

api_clients = {
    "github": GithubClient,
    "gitee": GiteeClient,
    "gitea": GiteaClient
}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lib', 'lib.api', 'lib.handler', 'lib.model']

package_data = \
{'': ['*']}

modules = \
['cli', 'pyproject']
install_requires = \
['docopt>=0.6.2,<0.7.0',
 'gitpython>=3.1.30,<4.0.0',
 'loguru>=0.6.0,<0.7.0',
 'pyyaml>=6.0,<7.0',
 'requests>=2.28.2,<3.0.0']

entry_points = \
{'console_scripts': ['trident = cli:cli']}

setup_kwargs = {
    'name': 'trident-sync',
    'version': '1.1.9',
    'description': '三叉戟，二次开发项目同步升级工具，Secondary development repo sync and upgrade CLI',
    'long_description': '# 🔱 trident-sync 三叉戟同步\n\n三叉戟同步，是一款异构项目同步升级工具，二次开发同步神器。\n\n[中文](./readme.md)\n\n<div>\n<a href="https://gitee.com/handsfree-work/trident-sync" target="_blank"\n  ><img src="https://gitee.com/handsfree-work/trident-sync/badge/star.svg?theme=dark" alt="star"\n/></a>  \n<a href="https://github.com/handsfree-work/trident-sync" target="_blank"\n  ><img alt="GitHub stars" src="https://img.shields.io/github/stars/handsfree-work/trident-sync?logo=github"\n/></a>\n</div>\n\n## 1. 简介\n\n当我们的项目内部使用了其他项目进行二次开发，那么这个模块就只能停留在当时的版本，无法方便的更新。\n\n本工具可以自动获取变更并合并到你的项目仓库，让二次开发项目可以持续升级。\n\n本工具适用于所有二次开发需要同步更新的场景：\n\n* 直接copy源项目进行二次开发。\n* 源项目与你项目目录结构不一致（异构）。\n* 源项目本身的submodule过多，一个个fork、merge太麻烦。\n* 源项目只是你项目的一个子模块，但你不想用submodule\n\n## 2. 缘起\n\n我有一个 [certd](https://github.com/certd/certd) 项目，这是一个自动更新ssl证书的工具，但这不是重点。     \n重点是它一开始只是一个独立的命令行工具。    \n目录结构如下：\n\n```\nsrc\n| --packages\n    | --core           \n    | --plugins        \n\n```\n\n某一天我想开发v2版本，把它升级成一个带后台和界面的web项目。      \n恰好我找到了两个模版项目(其实也是我写的🤭),可以帮我快速实现以上需求。\n\n* [fs-admin-antdv](https://github.com/fast-crud/fs-admin-antdv)  （前端admin模版）\n* [fs-server-js](https://github.com/fast-crud/fs-server-js)  （服务端）\n\n我把这两个项目复制到了`certd`项目中，进行二次开发。     \n此时`certd`项目目录结构变成如下：\n\n```\nsrc\n| --packages\n    | --core\n    | --plugins\n    | --ui\n        | --certd-client   //这是fs-admin-antdv的副本\n        | --certd-server   //这是fs-server-js的副本\n```\n\n为了使`certd-client`和`certd-server`能够随时同步`模版项目`的更新       \n我将使用本项目`trident-sync`来自动帮我升级。\n\n<p align="center">\n<img src="./doc/images/trident.png" height="400"/>\n<p align="center">像不像个三叉戟🔱？</p>\n<p>\n\n## 3. 原理过程\n\n初始化（init）：\n\n1. 初始化`同步工作仓库`（sync_work_repo）\n2. `clone` `源仓库`（src_repo）和`目标仓库`（target_repo），添加到`同步工作仓库`的`submodule`\n3. 给`目标仓库`创建并切换到`同步分支`（sync_branch）\n4. 将`源仓库内的文件`复制到`目标仓库对应的目录`，然后`commit、push`\n5. 此时`目标仓库`内的`sync_branch`分支拥有`源仓库`的副本\n\n同步（sync）：\n\n1. 当`源仓库`有变更时、拉取`源仓库`更新\n2. 删除`目标仓库对应的目录`，复制`源仓库所有文件`到`目标仓库对应的目录`\n3. 此时`git add . && git commit` 提交的就是`源仓库变更部分`\n4. 至此我们成功将`源仓库的更新`转化成了`目标仓库的commit`，后续就是常规的合并操作了。\n5. 创建`target.sync_branch` -> `target.main`的`PR`\n6. 处理`PR`，合并到开发主分支，升级完成\n\n<p align="center">\n<img src="./doc/images/desc.png" />\n<p align="center">同步流程图</p>\n<p>\n\n> 没有冲突的话，同步过程可以全部自动化。    \n> 解决冲突是唯一需要手动的部分。\n\n## 4. 快速开始\n\n### 4.1 准备工作\n\n* 安装 [python (3.8+)](https://www.python.org/downloads/)\n* 安装 `git`\n* 准备你的项目和要同步的源项目\n\n### 4.2 安装本工具\n\n```shell\n# 安装本工具，安装成功后就可以使用 trident 命令了\npip install trident-sync --upgrade\n```\n\n### 4.3 编写配置文件\n\n* 创建一个同步工作目录，你可以任意命名，接下来都在这个目录下进行操作\n\n```\nmkdir sync_work_repo\ncd sync_work_repo\n```\n\n* 编写`./sync_work_repo/sync.yaml`， 下面是示例，请根据其中注释说明改成你自己的内容\n\n```yaml\n# ./sync_work_repo/sync.yaml\nrepo: # 仓库列表，可以配置多个仓库\n  fs-admin: # 上游项目1，可以任意命名\n    url: "https://github.com/fast-crud/fs-admin-antdv" # 源仓库地址\n    path: "fs-admin-antdv"            # submodule保存路径，一般配置仓库名称即可\n    branch: "main"                    # 要同步过来的分支\n  certd: # 你的项目（接受同步项目），可以任意命名\n    url: "https://github.com/certd/certd"  # 目标仓库地址\n    path: "certd"                    # submodule保存路径，一般配置仓库名称即可\n    branch: "dev"                    # 你的代码开发主分支（接受合并的分支）例如dev、main、v1、v2等\n    # 以下配置与PR相关，更多关于PR的文档请前往 https://github.com/handsfree-work/trident-sync/tree/main/doc/pr.md\n    # 第一次使用，你可以暂时不配置，同步完之后需要手动操作合并\n    token: ""                         # 仓库的token，用于提交PR\n    type: github                      # 仓库类型，用于提交PR，可选项：[github/gitee/gitea]\n    auto_merge: true                  # 是否自动合并,如果有冲突则需要手动处理\n# 注意： 初始化之后，不要修改url和path，以免出现意外。但是可以添加新的repo.\n\nsync: # 同步配置，可以配置多个同步任务\n  client: # 同步任务1，可以任意命名\n    src: # 源仓库\n      repo: fs-admin                  # 源仓库名称，上面repo配置的仓库引用\n      dir: \'.\'                        # 要同步给target的目录（不能为空目录）\n    target: #接受合并的仓库，就是你的项目\n      repo: certd                     # 目标仓库名称，上面repo配置的仓库引用\n      dir: \'package/ui/certd-client\'  # 接收src同步过来的目录\n      # ↑↑↑↑↑↑↑↑↑↑↑↑↑↑↑↑↑↑↑↑↑↑↑↑↑↑↑↑    如果你之前已经使用源仓库副本做了一部分特性开发，那么这里配置源仓库副本的目录）\n      branch: \'client_sync\'           # 同步分支名称（需要配置一个未被占用的分支名称）\n\noptions: #其他选项\n  repo_root: repo          # submodule保存根目录\n  push: false              # 同步后是否push\n  # ↑↑↑↑↑↑↑↑↑↑↑↑↑↑↑↑↑↑↑↑↑↑↑  第一次使用，先本地测试，不push，没问题之后再改成true\n  pull_request: true       # 是否创建pull request，需要目标仓库配置token和type\n  proxy_fix: true          # 是否将https代理改成http://开头，解决python开启代理时无法发出https请求的问题\n  use_system_proxy: true   # 是否使用系统代理\n\n```\n\n### 4.4 初始化\n\n此命令会将`sync_work_repo`目录初始化成一个git仓库，然后将`sync.yaml`中配置的`repo` 添加为`submodule`\n\n```shell\ncd sync_work_repo\n# 执行初始化操作\ntrident init \n```\n\n<p align="center">\n<img src="./doc/images/init.png" />\n<p align="center">初始化执行效果</p>\n<p>\n\n> 只需运行一次即可，除非你添加了新的`repo`    \n\n> 初始化过程会将多个仓库添加为submodule，此步骤在网络不好时容易出问题   \n> 你可以删除目录下除`sync.yaml`之外的所有文件，重新运行`trident init`命令\n\n### 4.5 进行同步\n\n将根据`sync.yaml`中`sync`配置的同步任务进行同步更新，并提交PR，[你需要视情况处理PR](#5-合并分支)\n\n```shell\n# 以后你只需要定时运行这个命令，即可保持同步升级\ntrident sync \n```\n\n运行效果\n\n```\nroot:~/sync_work_repo$ trident sync \n.\n.\n.\n2023-01-28 14:13:41 | INFO    | - refs:[<git.Head "refs/heads/main">]\n2023-01-28 14:13:41 | WARNING | - Skip push，The remote address is not set for the current repository. Use the [trident remote <repo_url>] command to set the remote address of the repository and save the synchronization progress\n2023-01-28 14:13:41 | INFO    | - ----------------result:✅----------------\n 🏹 task            -->  success:✅     copy:✅   change:✅   commit:✅     push:✅       pr:✅    merge:✅\n 🔱 sync_work_repo  -->   change:✅   commit:✅     push:🚫 \n2023-01-28 14:13:41 | INFO    | - ----------------sync end----------------\n```\n\n> 注意：不要在同步分支内写你自己的任何代码（示例配置中为`client_sync`分支）\n\n### 4.6 [可选] 保存 sync_work_repo\n\n将`sync_work_repo`push到远程服务器，防止更换电脑丢失同步进度。    \n后续你只需要`clone` `sync_work_repo` ，然后直接运行`trident sync`即可继续同步\n\n```shell\n# 给同步仓库设置远程地址，并push\ntrident remote --url=<sync_work_repo_git_url> \n\n# 或者运行如下命令，一样的\ngit remote add origin <sync_work_repo_git_url> \ngit push\n```\n\n> 注意： `sync_work_repo_git_url` 应该是一个新的空的远程仓库     \n> 如果不是空的，可以加 `-f` 选项强制push（sync_work_repo原有的内容会被覆盖）。\n\n### 4.7 [可选] 定时运行\n\n你可以将 `<sync_work_repo>` 这个远程仓库和 `trident sync` 命令配置到任何`CI/DI`工具（例如jenkins、github\naction、drone等）自动定时同步\n\n## 5 合并分支\n\n源仓库如果有更新，那么同步完之后，将会有三种情况：\n\n* 启用了PR： [如何启用PR？](#启用PR)\n    * 无冲突：自动创建PR，然后自动合并，你无需任何操作\n    * 有冲突：自动创建PR，然后需要 [手动处理PR](#处理PR)\n* 未启用PR：\n    * 你需要 [手动合并](#手动合并)\n\n#### 启用PR\n\n要启用PR，你需要如下配置\n\n```yaml\nrepo:\n  target:\n    token: xxxx      # 创建PR的token\n    type: github     # upstream类型，支持[ github | gitee | gitea ]\n    auto_merge: true   # 是否自动合并\n\n```\n\n[token如何获取？](./doc/pr.md)\n\n#### 处理PR\n\n当PR有冲突时，就需要手动处理冲突，才能合并进入主分支\n\n* 其中 `github` `gitee`支持在web页面直接手动解决冲突\n* `gitea`需要线下解决，此时你仍然需要 [手动合并](#手动合并)\n\n#### 手动合并\n\n一般出现冲突了，都建议在IDE上手动进行合并\n\n1. 关闭PR（没有PR的话，请无视）\n2. 本地更新所有分支\n3. 通过IDE进行分支merge操作（rebase也行，用你平常熟悉的合并分支操作）\n\n```\ntarget:<sync_branch> -------->  target:<main_branch>\n    同步分支            merge         开发主分支\n```\n\n#### 避免冲突建议\n\n我们应该尽量避免冲突，请实际开发中遵循以下原则：\n\n1. 尽量不删除、不移动源项目的目录和文件（否则容易造成意想不到的难以解决的冲突）\n2. 尽量少在源项目的文件上进行修改（可以改，但尽量少）\n3. 新功能和新特性应该写在自己建立的新目录和新文件中\n\n总结就是六个字： 不删、少改、多加。\n\n## 6. 自动化\n\n### 6.1 GitHub Actions\nGitHub Actions 可以免费的帮你进行自动化同步操作\n\n请参考 [certd 自动化同步示例](https://github.com/certd/certd-sync-work)\n',
    'author': 'xiaojunnuo',
    'author_email': 'xiaojunnuo@qq.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'py_modules': modules,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4',
}


setup(**setup_kwargs)



from lib.api.abstract_client import AbstractClient


class GiteaClient(AbstractClient):
    '''gitea api'''
    '''
    http://gitea.baode.docmirror.cn/{repos}/{owner}
    
    https://try.gitea.io/api/swagger#/repository/repoMergePullRequest
    
    '''

    def __init__(self, http, token, url):
        super().__init__(http, token, url)
        self.api_root = f"{self.url_prefix}/api/v1"

    def post_pull_request(self, title, body, head_branch, base_branch):
        api = f"{self.api_root}/repos/{self.repo_path}/pulls?access_token={self.token}"
        res = self.http.post(api, data={
            "title": title,
            "body": body,
            # "head": f"{self.owner}:{src_branch}",
            "head": f"{head_branch}",
            "base": base_branch
        }, res_is_standard=False, res_is_json=True)
        return res

    def query_pull_request(self, head_branch, base_branch, state='open'):
        api = f"{self.api_root}/repos/{self.repo_path}/pulls?access_token={self.token}&head={head_branch}&base={base_branch}&state={state}"
        # api = f"{self.api_root}/repos/{self.repo_path}/pulls?access_token={self.token}&head={src_branch}&base={target_branch}&state=open"
        res = self.http.get(api, res_is_standard=False, res_is_json=True)
        if len(res) > 0:
            return res[0]
        return None

    def get_pull_request_detail(self, pull_id):
        api = f"{self.api_root}/repos/{self.repo_path}/pulls/{pull_id}?access_token={self.token}"
        res = self.http.get(api, res_is_standard=False, res_is_json=True)
        return res

    def post_merge(self, pull_id, detail):
        api = f"{self.api_root}/repos/{self.repo_path}/pulls/{pull_id}/merge?access_token={self.token}"
        self.http.post(api, data={
            "Do": 'merge'
        }, res_is_standard=False, res_is_json=False)



from sys import stdout

import loguru

logger = loguru.logger
logger.remove(0)
logger.add(stdout, format="<green>{time:YYYY-MM-DD HH:mm:ss}</green> | <level>{level:7}</level> | - <level>{message}</level>", colorize=True)
